@@ -1,10 +1,10 @@
 /*
     aviplay.c
-    copyright (c) 1998-2006 Kazuki IWAMOTO http://www.maid.org/ iwm@maid.org
+    copyright (c) 1998-2019 Kazuki Iwamoto https://www.maid.org/ iwm@maid.org
 
-    This program is free software; you can redistribute it and/or modify
+    This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
+    the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
@@ -13,8 +13,7 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 #include <ctype.h>
 #include <stdio.h>
@@ -24,7 +23,7 @@
 
 
 #ifndef min
-#define min(a,b) (((a)<(b))?(a):(b))
+# define min(a,b) (((a)<(b))?(a):(b))
 #endif /* not min */
 
 
@@ -50,11 +49,19 @@ get_value (const unsigned char *buf,
            const int            offset,
            const size_t         size)
 {
-  int i;
   long value = 0;
 
-  for (i = 0; i < size; i++)
-    value += (long)buf[i + offset] << (i * 8);
+  if (size > 0)
+    {
+      const unsigned char *p;
+      int i;
+
+      p = buf + offset + size - 1;
+      if ((char)*p < 0)
+        value = -1;
+      for (i = 0; i < size; i++)
+        value = value << 8 | *p--;
+    }
   return value;
 }
 
@@ -82,6 +89,7 @@ main (int   argc,
   int scale, rate;          /* rate÷scale=1秒間のフレーム数 */
   int bwidth, bheight;      /* ビットマップのサイズ */
   int bcount;               /* ビットマップのビット数 */
+  int topdown = 0;          /* ビットマップのトップダウン */
   long offset;              /* moviフォームのLISTチャンクのオフセット */
   size_t image;             /* ピクセルデータのバイト数 */
   size_t line;              /* 横ラインのバイト数 */
@@ -92,7 +100,7 @@ main (int   argc,
   FILE *fp2;                /* イメージデータを読むためのファイルポインタ */
 
   for (i = 1; i < argc; i++)
-    if ((argv[i])[0] != '-' && (argv[i])[0] != '/')
+    if ((argv[i])[0] != '-')
       /* オプションではなくファイル名 */
       s = i;
     else
@@ -105,7 +113,7 @@ main (int   argc,
           case 'L':/* 明るい背景に暗い文字 */
             text = "%8&s0x67to=+-:. ";
             break;
-          case 'S':/* 明るい背景に暗い文字 */
+          case 'S':/* フレーム落ちしているときには表示をスキップ */
             skip = atoi (argv[i] + 2);
             if (skip <= 0)
               skip = 1;
@@ -120,7 +128,7 @@ main (int   argc,
     {
       /* 引数にファイル名がないとき */
       fprintf (stderr,
-"copyright (c) 1998-2006 Kazuki IWAMOTO http://www.maid.org/ iwm@maid.org\n"
+"copyright (c) 1998-2019 Kazuki Iwamoto https://www.maid.org/ iwm@maid.org\n"
 "This program comes with NO WARRANTY, to the extent permitted by law.\n"
 "You may redistribute it under the terms of the GNU General Public License;\n"
 "see the file named COPYING for details.\n"
@@ -150,13 +158,13 @@ main (int   argc,
   ck = chunkopen (fp1);
   if (!ck)
     {
-      fprintf (stderr, "chunk open error.\n");
+      fputs ("chunk open error.\n", stderr);
       return -1;
     }
   /* RIFFチャンクに入る */
   if (chunkin (fp1, ck) != 0)
     {
-      fprintf (stderr, "RIFF\'AVI \' chunk in error.\n");
+      fputs ("RIFF\'AVI \' chunk in error.\n", stderr);
       chunkfree (ck);
       return -1;
     }
@@ -164,14 +172,14 @@ main (int   argc,
   while (chunkform (fp1) != make4cc ('h', 'd', 'r', 'l'))
     if (chunknext (fp1, ck) != 0)
       {
-        fprintf (stderr, "RIFF\'AVI \' -> LIST\'hdrl\' chunk search error.\n");
+        fputs ("RIFF\'AVI \' -> LIST\'hdrl\' chunk search error.\n", stderr);
         chunkfree (ck);
         return -1;
       }
   /* フォームがhdrlのLISTチャンクに入る(RIFF'AVI ' -> LIST'hdrl') */
   if (chunkin (fp1, ck) != 0)
     {
-      fprintf (stderr, "RIFF\'AVI \' -> LIST\'hdrl\' chunk in error.\n");
+      fputs ("RIFF\'AVI \' -> LIST\'hdrl\' chunk in error.\n", stderr);
       chunkfree (ck);
       return -1;
     }
@@ -179,8 +187,8 @@ main (int   argc,
   while (chunkform (fp1) != make4cc ('s', 't', 'r', 'l'))
     if (chunknext (fp1, ck) != 0)
       {
-        fprintf (stderr, "RIFF\'AVI \' -> LIST\'hdrl\' -> LIST\'strl\' "
-                                                    "chunk search error.\n");
+        fputs ("RIFF\'AVI \' -> LIST\'hdrl\' -> LIST\'strl\' "
+                                            "chunk search error.\n", stderr);
         chunkfree (ck);
         return -1;
       }
@@ -188,8 +196,8 @@ main (int   argc,
                                     (RIFF'AVI ' -> LIST'hdrl' -> LIST'strl') */
   if (chunkin (fp1, ck) != 0)
     {
-      fprintf (stderr, "RIFF\'AVI \' -> LIST\'hdrl\' -> LIST\'strl\' "
-                                                        "chunk in error.\n");
+      fputs ("RIFF\'AVI \' -> LIST\'hdrl\' -> LIST\'strl\' "
+                                                "chunk in error.\n", stderr);
       chunkfree (ck);
       return -1;
     }
@@ -197,9 +205,8 @@ main (int   argc,
   while (chunkid (fp1) != make4cc ('s', 't', 'r', 'h'))
     if (chunknext (fp1, ck) != 0)
       {
-        fprintf (stderr,
-                        "RIFF\'AVI \' -> LIST\'hdrl\' -> LIST\'strl\' -> strh "
-                                                    "chunk search error.\n");
+        fputs ("RIFF\'AVI \' -> LIST\'hdrl\' -> LIST\'strl\' -> strh "
+                                            "chunk search error.\n", stderr);
         chunkfree (ck);
         return -1;
       }
@@ -207,14 +214,14 @@ main (int   argc,
                             (RIFF'AVI ' -> LIST'hdrl' -> LIST'strl' -> strh) */
   if (chunkin (fp1, ck) != 0)
     {
-      fprintf (stderr, "RIFF\'AVI \' -> LIST\'hdrl\' -> LIST\'strl\' -> strh "
-                                                        "chunk in error.\n");
+      fputs ("RIFF\'AVI \' -> LIST\'hdrl\' -> LIST\'strl\' -> strh "
+                                                "chunk in error.\n", stderr);
       chunkfree (ck);
       return -1;
     }
   if (fread (&ash, sizeof (AVISTREAMHEADER), 1, fp1) != 1)
     {
-      fprintf (stderr, "AVISTREAMHEADER read error.\n");
+      fputs ("AVISTREAMHEADER read error.\n", stderr);
       chunkfree (ck);
       return -1;
     }
@@ -222,14 +229,14 @@ main (int   argc,
   rate = ash_rate (&ash);
   if (ash_type (&ash) != make4cc ('v', 'i', 'd', 's'))
     {
-      fprintf (stderr, "first stream is not video.\n");
+      fputs ("first stream is not video.\n", stderr);
       chunkfree (ck);
       return -1;
     }
   if (ash_handler (&ash) != make4cc ('D', 'I', 'B', ' ')
                                                     && ash_handler (&ash) != 0)
     {
-      fprintf (stderr, "first stream is compressed.\n");
+      fputs ("first stream is compressed.\n", stderr);
       chunkfree (ck);
       return -1;
     }
@@ -237,8 +244,8 @@ main (int   argc,
                             (RIFF'AVI ' -> LIST'hdrl' -> LIST'strl' <- strh) */
   if (chunkout (fp1, ck) != 0)
     {
-      fprintf (stderr, "RIFF\'AVI \' -> LIST\'hdrl\' -> LIST\'strl\' <- strh "
-                                                        "chunk out error.\n");
+      fputs ("RIFF\'AVI \' -> LIST\'hdrl\' -> LIST\'strl\' <- strh "
+                                                "chunk out error.\n", stderr);
       chunkfree (ck);
       return -1;
     }
@@ -246,8 +253,8 @@ main (int   argc,
                                     (RIFF'AVI ' -> LIST'hdrl' <- LIST'strl') */
   if (chunkout (fp1, ck) != 0)
     {
-      fprintf (stderr, "RIFF\'AVI \' -> LIST\'hdrl\' <- LIST\'strl\' "
-                                                        "chunk out error.\n");
+      fputs ("RIFF\'AVI \' -> LIST\'hdrl\' <- LIST\'strl\' "
+                                                "chunk out error.\n", stderr);
       chunkfree (ck);
       return -1;
     }
@@ -255,8 +262,8 @@ main (int   argc,
                                     (RIFF'AVI ' -> LIST'hdrl' -> LIST'strl') */
   if (chunkin (fp1, ck) != 0)
     {
-      fprintf (stderr, "RIFF\'AVI \' -> LIST\'hdrl\' -> LIST\'strl\' "
-                                                        "chunk in error.\n");
+      fputs ("RIFF\'AVI \' -> LIST\'hdrl\' -> LIST\'strl\' "
+                                                "chunk in error.\n", stderr);
       chunkfree (ck);
       return -1;
     }
@@ -264,9 +271,8 @@ main (int   argc,
   while (chunkid (fp1) != make4cc ('s', 't', 'r', 'f'))
     if (chunknext (fp1, ck) != 0)
       {
-        fprintf (stderr,
-                        "RIFF\'AVI \' -> LIST\'hdrl\' -> LIST\'strl\' -> strf "
-                                                    "chunk search error.\n");
+        fputs ("RIFF\'AVI \' -> LIST\'hdrl\' -> LIST\'strl\' -> strf "
+                                            "chunk search error.\n", stderr);
         chunkfree (ck);
         return -1;
       }
@@ -274,24 +280,29 @@ main (int   argc,
                             (RIFF'AVI ' -> LIST'hdrl' -> LIST'strl' -> strf) */
   if (chunkin (fp1, ck) != 0)
     {
-      fprintf (stderr, "RIFF\'AVI \' -> LIST\'hdrl\' -> LIST\'strl\' -> strf "
-                                                        "chunk in error.\n");
+      fputs ("RIFF\'AVI \' -> LIST\'hdrl\' -> LIST\'strl\' -> strf "
+                                                "chunk in error.\n", stderr);
       chunkfree (ck);
       return -1;
     }
   /* BITMAPINFOHEADERを読み込む */
   if (fread (&bmih, sizeof (BITMAPINFOHEADER), 1, fp1) != 1)
     {
-      fprintf (stderr, "BITMAPINFOHEADER can not read.\n");
+      fputs ("BITMAPINFOHEADER can not read.\n", stderr);
       chunkfree (ck);
       return -1;
     }
   bwidth = bmih_width (&bmih);
   bheight = bmih_height (&bmih);
+  if (bheight < 0)
+    {
+      bheight = -bheight;
+      topdown = 1;
+    }
   bcount = bmih_bit_count (&bmih);
   /* チャンクポインタを解放する */
   if (chunkfree (ck) != 0)
-    fprintf (stderr, "chunk free error.\n");
+    fputs ("chunk free error.\n", stderr);
 
   /* ムービーチャンクへ移動する */
 
@@ -299,13 +310,13 @@ main (int   argc,
   ck = chunkopen (fp1);
   if (!ck)
     {
-      fprintf (stderr, "chunk open error.\n");
+      fputs ("chunk open error.\n", stderr);
       return -1;
     }
   /* RIFFチャンクに入る */
   if (chunkin (fp1, ck) != 0)
     {
-      fprintf (stderr, "RIFF\'AVI \' chunk in error.\n");
+      fputs ("RIFF\'AVI \' chunk in error.\n", stderr);
       chunkfree (ck);
       return -1;
     }
@@ -313,7 +324,7 @@ main (int   argc,
   while (chunkform (fp1) != make4cc ('m', 'o', 'v', 'i'))
     if (chunknext (fp1, ck) != 0)
       {
-        fprintf (stderr, "RIFF\'AVI \' -> LIST\'movi\' chunk search error.\n");
+        fputs ("RIFF\'AVI \' -> LIST\'movi\' chunk search error.\n", stderr);
         chunkfree (ck);
         return -1;
       }
@@ -321,7 +332,7 @@ main (int   argc,
   offset = ftell (fp1);
   if (offset == -1)
     {
-      fprintf (stderr, "file pointer error.\n");
+      fputs ("file pointer error.\n", stderr);
       chunkfree (ck);
       return -1;
     }
@@ -329,7 +340,7 @@ main (int   argc,
                 + sizeof (cksize_t) + sizeof (fourcc_t);/* moviのオフセット */
   /* チャンクポインタを解放する */
   if (chunkfree (ck) != 0)
-    fprintf (stderr, "chunk free error.\n");
+    fputs ("chunk free error.\n", stderr);
 
   /* インデックスチャンクへ移動する */
 
@@ -337,13 +348,13 @@ main (int   argc,
   ck = chunkopen (fp1);
   if (!ck)
     {
-      fprintf (stderr, "chunk open error.\n");
+      fputs ("chunk open error.\n", stderr);
       return -1;
     }
   /* RIFFチャンクに入る */
   if (chunkin (fp1, ck) != 0)
     {
-      fprintf (stderr, "RIFF\'AVI \' chunk in error.\n");
+      fputs ("RIFF\'AVI \' chunk in error.\n", stderr);
       chunkfree (ck);
       return -1;
     }
@@ -351,7 +362,7 @@ main (int   argc,
   while (chunkid (fp1) != make4cc ('i', 'd', 'x', '1'))
     if (chunknext (fp1, ck) != 0)
       {
-        fprintf (stderr, "RIFF\'AVI \' -> idx1 chunk search error.\n");
+        fputs ("RIFF\'AVI \' -> idx1 chunk search error.\n", stderr);
         chunkfree (ck);
         return -1;
       }
@@ -359,14 +370,14 @@ main (int   argc,
   size = chunksize (fp1);
   if (size == (cksize_t)-1)
     {
-      fprintf (stderr, "idx1 chunk size error.\n");
+      fputs ("idx1 chunk size error.\n", stderr);
       chunkfree (ck);
       return -1;
     }
   /* idx1チャンクに入る(RIFF'AVI ' -> idx1) */
   if (chunkin (fp1, ck) != 0)
     {
-      fprintf (stderr, "RIFF\'AVI \' -> idx1 chunk in error.\n");
+      fputs ("RIFF\'AVI \' -> idx1 chunk in error.\n", stderr);
       chunkfree (ck);
       return -1;
     }
@@ -423,7 +434,7 @@ main (int   argc,
   q[(cx + 1) * cy] = '\0';
   if (!p)
     {
-      fprintf (stderr, "memory alloc error.\n");
+      fputs ("memory alloc error.\n", stderr);
       chunkfree (ck);
       return -1;
     }
@@ -433,17 +444,15 @@ main (int   argc,
   t = time (NULL);
   for (i = 0; i < s; i++)
     {
-      size_t length;
       AVIINDEXENTRY aie;        /* AVIINDEXENTRY構造体 */
 
       if (fread (&aie, sizeof (AVIINDEXENTRY), 1, fp1) != 1)
         {
-          fprintf (stderr, "AVIINDEXENTRY can not read.\n");
+          fputs ("AVIINDEXENTRY can not read.\n", stderr);
           break;
         }
-      length = aie_length (&aie);
       if (aie_ckid (&aie) == make4cc ('0', '0', 'd', 'b')
-                                        && (length == 0 || length == image))
+                            || aie_ckid (&aie) == make4cc ('0', '0', 'd', 'c'))
         {
           clock_t clk;
           time_t tm;
@@ -457,66 +466,109 @@ main (int   argc,
           if (frame % skip != 0
                 || (double)frame * CLOCKS_PER_SEC * scale / rate <= clk - c)
             continue;/* フレーム落ちしているときには表示をスキップ */
-          if (length > 0)
+          if (aie_length (&aie) >= image)
             {
               int x, y;
 
               /* ピクセルデータへシークする */
               if (fseek (fp2, offset + aie_offset (&aie), SEEK_SET) != 0)
                 {
-                  fprintf (stderr, "file seek error.\n");
+                  fputs ("file seek error.\n", stderr);
                   break;
                 }
               /* ピクセルデータを読み込む */
-              if (fread (p, sizeof (unsigned char), length, fp2) != length)
+              if (fread (p, sizeof (unsigned char), image, fp2) != image)
                 {
-                  fprintf (stderr, "pixel data can not read.\n");
+                  fputs ("pixel data can not read.\n", stderr);
                   break;
                 }
               switch (bcount)
                 {
                   case 16:
-                    for (y = 0; y < cy; y++)
-                      for (x = 0; x < cx; x++)
-                        {
-                          int r, g, b;
-                          long adr;                 /* ピクセルの位置 */
+                    if (topdown == 0)
+                      for (y = 0; y < cy; y++)
+                        for (x = 0; x < cx; x++)
+                          {
+                            int r, g, b;
+                            long adr;               /* ピクセルの位置 */
 
-                          adr = bheight * (cy - y - 1) / cy * line
+                            adr = bheight * (cy - y - 1) / cy * line
                                                         + bwidth * x / cx * 2;
-                          b = (p[adr] & 0x1f);
-                          g = ((p[adr] & 0xe0) >> 5)
+                            b = (p[adr] & 0x1f);
+                            g = ((p[adr] & 0xe0) >> 5)
                                                 + ((p[adr + 1] & 0x03) << 3);
-                          r = (p[adr + 1] & 0x7c) >> 2;
-                          q[x + y * (cx + 1)] = text[(b * 29 + g * 150
+                            r = (p[adr + 1] & 0x7c) >> 2;
+                            q[x + y * (cx + 1)] = text[(b * 29 + g * 150
                                                             + r * 77) / 512];
-                        }
+                          }
+                    else
+                      for (y = 0; y < cy; y++)
+                        for (x = 0; x < cx; x++)
+                          {
+                            int r, g, b;
+                            long adr;               /* ピクセルの位置 */
+
+                            adr = bheight * y / cy * line
+                                                        + bwidth * x / cx * 2;
+                            b = (p[adr] & 0x1f);
+                            g = ((p[adr] & 0xe0) >> 5)
+                                                + ((p[adr + 1] & 0x03) << 3);
+                            r = (p[adr + 1] & 0x7c) >> 2;
+                            q[x + y * (cx + 1)] = text[(b * 29 + g * 150
+                                                            + r * 77) / 512];
+                          }
                     break;
                   case 24:
-                    for (y = 0; y < cy; y++)
-                      for (x = 0; x < cx; x++)
-                        {
-                          long adr;                 /* ピクセルの位置 */
+                    if (topdown == 0)
+                      for (y = 0; y < cy; y++)
+                        for (x = 0; x < cx; x++)
+                          {
+                            long adr;               /* ピクセルの位置 */
 
-                          adr = bheight * (cy - y - 1) / cy * line
+                            adr = bheight * (cy - y - 1) / cy * line
                                                         + bwidth * x / cx * 3;
-                          q[x + y * (cx + 1)] = text[(p[adr] * 29
+                            q[x + y * (cx + 1)] = text[(p[adr] * 29
                                                     + p[adr + 1] * 150
                                                     + p[adr + 2] * 77) / 4096];
-                        }
+                          }
+                    else
+                      for (y = 0; y < cy; y++)
+                        for (x = 0; x < cx; x++)
+                          {
+                            long adr;               /* ピクセルの位置 */
+
+                            adr = bheight * y / cy * line
+                                                        + bwidth * x / cx * 3;
+                            q[x + y * (cx + 1)] = text[(p[adr] * 29
+                                                    + p[adr + 1] * 150
+                                                    + p[adr + 2] * 77) / 4096];
+                          }
                     break;
                   case 32:
-                    for (y = 0; y < cy; y++)
-                      for (x = 0; x < cx; x++)
-                        {
-                          long adr;                 /* ピクセルの位置 */
+                    if (topdown == 0)
+                      for (y = 0; y < cy; y++)
+                        for (x = 0; x < cx; x++)
+                          {
+                            long adr;               /* ピクセルの位置 */
 
-                          adr = bheight * (cy - y - 1) / cy * line
+                            adr = bheight * (cy - y - 1) / cy * line
                                                         + bwidth * x / cx * 4;
-                          q[x + y * (cx + 1)] = text[(p[adr] * 29
+                            q[x + y * (cx + 1)] = text[(p[adr] * 29
                                                     + p[adr + 1] * 150
                                                     + p[adr + 2] * 77) / 4096];
-                        }
+                          }
+                    else
+                      for (y = 0; y < cy; y++)
+                        for (x = 0; x < cx; x++)
+                          {
+                            long adr;               /* ピクセルの位置 */
+
+                            adr = bheight * y / cy * line
+                                                        + bwidth * x / cx * 4;
+                            q[x + y * (cx + 1)] = text[(p[adr] * 29
+                                                    + p[adr + 1] * 150
+                                                    + p[adr + 2] * 77) / 4096];
+                          }
                 }
               if (esc != 0 && i > 0)
                 printf ("\x1b[%dA%s", cy, q);/* エスケープシーケンスで移動 */
@@ -529,7 +581,7 @@ main (int   argc,
     }
   /* チャンクポインタを解放する */
   if (chunkfree (ck) != 0)
-    fprintf (stderr, "chunk free error.\n");
+    fputs ("chunk free error.\n", stderr);
   free (p);
   free (q);
   return 0;
